import numpy as np
from astropy.table import Table, Column
from edge_pydb.conversion import gc_polr
from functools import wraps

use_numba = False

if use_numba:
    use_python = False
    try: 
        from numba import jit, njit, prange
        from numba.typed import List
    except (ImportError, ModuleNotFoundError):
        print("cannot import numba modules, computation will be slow")
        use_python = True
else:
    use_python = True

if use_python:
    prange = range
    List = list
    def py_njit(parallel=False):
        def py_njit_func(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                retval = func(*args, **kwargs)
                return retval
            return wrapper
        return py_njit_func
    
    def py_jit(func):
        return func

    jit = py_jit
    njit = py_njit

@jit
def ylin_hex(pos, hex_sidelen, bound):
    '''
    going along y dir
    '''
#     print(pos, hex_sidelen, bound)
    dist = np.sqrt(3) * hex_sidelen
    y_plus = np.arange(pos[1], bound[1][1], dist)
    num = np.floor((pos[1] - bound[0][1]) / dist)
    y_minus = np.arange(pos[1] - num * dist, pos[1], dist)
    y_coord = np.concatenate([y_minus, y_plus])
    return np.array(list(zip(np.ones_like(y_coord) * pos[0], y_coord)))

@jit
def hex_basis(ref, hex_sidelen, bound):
    dist = np.sqrt(3) * hex_sidelen
    angle = np.pi/6
    delta = np.array((dist * np.cos(angle), dist * np.sin(angle)))
    basis_pos = []
    ref = np.array((float(ref[0]), float(ref[1])))
    plus = np.copy(ref)
#     print(bound[1][0])
    while plus[0] < bound[1][0] and plus[1] < bound[1][1]:
        basis_pos.append(np.copy(plus))
#         print(plus)
        plus += delta
    minus = np.copy(ref) - delta
    while minus[0] > bound[0][0] and minus[1] > bound[0][1]:
        basis_pos.append(np.copy(minus))
        minus -= delta
        
    return np.array(basis_pos)

@jit
def hex_grid(ref, sidelen, bound, starting_angle, precision):
    # hex side length same as the circumcircle Radius, = 2 / \sqrt(3) * incircle radius
    x_len = bound[1][0] - bound[0][0]
    y_len = bound[1][1] - bound[0][1]
    if x_len > y_len:
        max_len = x_len
    else:
        max_len = y_len

    # create a maximum square 
    rotate_compen = np.array((max_len * (1 + np.cos(starting_angle)), max_len * (1 + np.cos(starting_angle))))
    max_bound = [bound[0] - rotate_compen, 
                 bound[0] + rotate_compen]

    x_basis = hex_basis(ref, sidelen, max_bound)
    grid = None
    for point in x_basis:
        tmp = ylin_hex(point, sidelen, max_bound)
        if grid is None:
            grid = np.copy(tmp)
        else:
            grid = np.concatenate((grid, tmp), axis=0)
    rotation_mtx = np.array([[np.cos(starting_angle), np.sin(starting_angle)], 
                             [-np.sin(starting_angle), np.cos(starting_angle)]])
    grid = np.dot(grid, rotation_mtx)
    cut = (grid[:, 0] > bound[0, 0]) & (grid[:, 1] > bound[0, 1]) \
     & (grid[:, 0] < bound[1, 0]) & (grid[:, 1] < bound[1, 1])
    # There will be repeated data points if we don't remove them
    # not_repeated = np.full(len(grid[cut]), False)
    # for i, coord in enumerate(grid[cut]):
    #     occurence = 0
    #     for to_compare in grid[cut]:
    #         if np.abs(coord[0] - to_compare[0]) < 1e-5 and np.abs(coord[1] - to_compare[1]) < 1e-5:
    #             occurence += 1
    #     if occurence <= 1:
    #         not_repeated[i] = True
    not_repeated = [True if np.where((np.abs(coord - grid) < 1e-5).all(axis=1))[0].shape[0] <= 1 else False for coord in grid[cut]]
    repeated_x = []
    repeated_y = []
    for i in range(len(not_repeated)):
        if not_repeated[i] == False \
        and grid[cut][i][0] not in repeated_x \
        and grid[cut][i][1] not in repeated_y:
            repeated_x.append(grid[cut][i][0])
            repeated_y.append(grid[cut][i][1])
            not_repeated[i] = True
    if precision != 0:
        grid = np.around(grid, precision)
    return grid[cut][not_repeated]

@jit
def interpolate_neighbor(point, bound, step_size=0):
    '''
    find the adjacent pixels of the point to interpolate from
    if the point is directly on one of the pixel, we take that value directly
    '''
    if step_size == 0:
        tmp = List([np.ceil(point), np.floor(point),
                np.array([np.floor(point[0]), np.ceil(point[1])]),
                np.array([np.ceil(point[0]), np.floor(point[1])])])
    else:
        i = -1 * step_size
        j = i
        tmp = List()
        cen = np.ceil(point)
        while i != step_size:
            while j != step_size:
                tmp.append(np.array([cen[0]+i, cen[1]+j]))
#                 print(np.array([cen[0]+i, cen[1]+j]))
                j += 1
            i += 1
    flag = False
    for k in range(len(tmp)): 
        coord = tmp[k]
        if bound[0][0] <= coord[0] <= bound[1][0] and bound[0][1] <= coord[1] <= bound[1][1]:
            if not flag:
#                 retval.append(coord)
                retval = np.reshape(coord, (-1, 2))
                flag = True
            else:
                retval = np.concatenate((retval, np.reshape(coord, (-1, 2))))
    return retval

# note that if useing numba, there might be an error caused by mixed use of float32 and float64
# when casting float64 to float32, memory can overflow.
@njit(parallel=True)
def interpolate_all_points(tab, datapoint, bound, header):
    sampled_tab = np.zeros((datapoint.shape[0], len(header)), dtype=np.float32)
    for j in prange(datapoint.shape[0]):
        coord = datapoint[j]
        inter = interpolate_neighbor(coord, bound)
        weight_arr = np.zeros(inter.shape[0])
        on_original_pix = False
        idx = 0 
        for i, point in enumerate(inter):
            cur = tab[(tab['ix'] == point[0]) & (tab['iy']==point[1])]
            distance = np.linalg.norm(point - coord)
            if distance == 0:
                idx = i
                on_original_pix = True
                break
            else:
                weight_arr[i] = 1. / distance
        # use a weight for each column
        if on_original_pix:
            row = tab[(tab['ix'] == inter[idx][0]) & (tab['iy']==inter[idx][1])]
        else:
            w = np.ones((len(weight_arr), len(header[2:])))
            for i in np.arange(len(weight_arr)):
                w[i, :] *= np.float32(weight_arr[i] / np.sum(weight_arr))
            row = np.zeros(len(header[2:]))
            for i in np.arange(inter.shape[0]):
                # talking about the Nans in the paper
                cur = tab[(tab['ix'] == inter[i][0]) & (tab['iy']==inter[i][1])]
                row =  np.sum([row, w[i, :] * cur.view(np.float32)[2:]], axis=0)
        sampled_tab[j, :2] = coord
        if type(row[0]) is np.void:
            row = np.array([np.float32(row[0][i]) for i in np.arange(2, len(row[0]))])
        sampled_tab[j, 2:] = row
#         for i in range(len(header[2:])):
#             if type(row[i]) is np.void:
#                 row = row[0]
#             # some edge cases
#             if row[i] == 0:
#                 sampled_tab[j, 2+i] = np.nan
#             else:
#                 sampled_tab[j, 2+i] = row[i]
    return sampled_tab

# @jit
def hex_sampler(tab, sidelen, keepref, ref_pix, ra_ref, dec_ref, 
                ra_gc, dec_gc, pa, inc,
                starting_angle=0, precision=2, hexgrid_output=None):
    '''
    a wrapper function to generate the hex grid and then interpolate and output the sampled data in 
    Astropy Table
    precision = 0 -> full precision
    hexgrid_output -> output file to check
    '''
    header = tab.colnames
    upper = np.array([float(max(tab[header[0]])), float(max(tab[header[1]]))])
    lower = np.array([float(min(tab[header[0]])), float(min(tab[header[1]]))])
    bound = np.array([lower, upper])
    if keepref:
        cen = ref_pix
    else:
        cen = np.array([80.,80.])
    # create a hex grid map
    datapoint= hex_grid(cen, sidelen, bound, starting_angle, precision)
    if hexgrid_output is not None:
        np.savetxt(hexgrid_output, datapoint)
    info = interpolate_all_points(tab.as_array(), datapoint, bound, List(header))
    units = [tab[col].unit for col in tab.colnames]
    sampled_tab = Table(info, names=header, units=units)
    if ('rad_arc' in header) and ('azi_ang' in header):
        sampled_tab['rad_arc'], sampled_tab['azi_ang'] = gc_polr(
                sampled_tab['ra_off'] + ra_ref,
                sampled_tab['dec_off'] + dec_ref,
                ra_gc, dec_gc, pa, inc
            )
    return sampled_tab



import os as _os
import json as _json
import shutil as _shutil
import requests as _requests
import h5py as _h5py


# Initial setup script to read the file location from config file
_ROOT = _os.path.abspath(_os.path.dirname(__file__))

_filepath = _os.path.join(_ROOT, '_config.json')
_runtime = False
_config = {}

try:
    _fp = open(_filepath, 'r')
    if _os.stat(_filepath).st_size > 0:
        _config = _json.load(_fp)

except FileNotFoundError:
    _fp = open(_filepath, 'w')
    # _config = {}

except OSError as _err:
    if _err.errno == 30:
        print("WARNING! Read-only file system, cannot save the package data file location.\n" +
              "For better longterm performance, consider providing a config file by using the extConfig() function.")
        print("If you need to change the files in the package data, please consider running as root, \
        the manipulation of files requires the sudo priority.")
    _runtime = True
    # _config = {}


def _walkthrough(dir=_ROOT, max_depth=2):
    retval = {}
    dir = _os.path.abspath(dir)
    base_depth = dir.rstrip(_os.path.sep).count(_os.path.sep)
    for _root, _dirs, _files in _os.walk(dir):
        cur_depth = _root.count(_os.path.sep)
        if max_depth >= 0 and cur_depth > max_depth + base_depth:
            continue
        for _file in _files:
            if _file.endswith('.csv') or _file.endswith('.hdf5'):
                if _file in retval:
                    print("{} redundant file detected\n--Current location: {}\n++New location: {}\n".format(
                        _file, retval[_file], _os.path.join(_root, _file)))
                retval[_file] = _os.path.join(_root, _file)
    return retval


if not _config:
    # print(os.listdir(_ROOT))
    _config = _walkthrough()
    if not _runtime:
        _json.dump(_config, _fp)

if not _runtime:
    _fp.close()



# update the files
def updatefiles(dir=_ROOT, max_depth=-1):
    tmp = _walkthrough(dir, max_depth)
    for k, v in tmp.items():
        if k not in _config.keys():
            print("Add file %s" % k)
        elif v != _config[k]:
            print("Update file %s" % k)
    _config.update(tmp)
    if not _runtime:
        with open(_filepath, 'w') as _fp:
            _json.dump(_config, _fp)


def download(file, url, loc='', user='', password=''):
    if not loc:
        if not _os.path.exists(_ROOT + '/data'):
            _os.mkdir(_ROOT + '/data')
        loc = _ROOT + '/data/'
    data = _requests.get(url + file, verify=False, auth=(user, password))
    if data.status_code != 200:
        data.raise_for_status()
    if not loc.endswith('/'):
        loc += '/'
    with open(loc + file, "wb") as _fp:
        _fp.write(data.content)


def save_config(src):
    '''
    This function will write _config back to a file.
    '''
    global _config
    with open(src, 'w') as _fp:
        _json.dump(_config, _fp)


def load_config(src, readonly=False):
    '''
    read the config from a file
    '''
    _fp = open(src, 'r')
    global _config
    if readonly:
        _config = _json.load(_fp)
    else:
        _config.update(_json.load(_fp))
        _filepath = src
        _runtime = False
    _fp.close()


def listfiles(contain='', values=False, printing=False):
    '''
    List the current available files in the package data directory
    If values=True, give full directory paths
    If contain='hdf', list only file names with the 'hdf' substring

    Parameters:
        contain: the target substring to find in the file name to list, if not provided, then will print all files
    '''
    files = []
    if values:
        for val in _config.values():
            if contain in val:
                if printing:
                    print(val)
                files.append(val)
    else:
        for key in _config.keys():
            if contain in key:
                if printing:
                    print(key)
                files.append(key)
    return files


def fetch(names):
    '''
    Get all the files by its file name, can either be a single file or a list of files

    Parameters:
        names: a list of files or a single file name in package data
    '''
    if isinstance(names, list):
        retval = []
        for name in names:
            if name not in _config.keys():
                raise FileNotFoundError("Cannot find the specified file: %s" % name)
            # if dir:
            #     dirpath = _os.path.abspath(_os.path.dirname(name))
            #     if dirpath not in retval:
            #         retval.append(path)
            else:
                retval.append(_config[name])
        return retval
    else:
        if names not in _config.keys():
            raise FileNotFoundError("Cannot find the specified file: %s" % names)
        # if dir:
        #     return _os.path.abspath(_os.path.dirname(names))
        else:
            return _config[names]


def addfile(src, dest='', copy=True, overwrite=False):
    if _runtime:
        print("WARNING! No sudo permission, take care, will break")
    name = _os.path.basename(src)
    src = _os.path.abspath(src)
    if copy:
        if dest:
            if name in _config.keys():
                if overwrite:
                    _os.remove(_config[name])
                else:
                    raise FileExistsError(_config[name])
            _shutil.copyfile(src, dest)
            _config[name] = dest
        else:
            if name in _config.keys():
                if overwrite:
                    _shutil.copyfile(src, _config[name])
                else:
                    raise FileExistsError('%s exists: ' % name + _config[name])

            if not _os.path.exists(_ROOT + '/data'):
                _os.mkdir(_ROOT + '/data')
            _shutil.copyfile(src, _ROOT + '/data/' + name)
            _config[name] = _ROOT + '/data/' + name
    else:
        if name in _config.keys():
            if overwrite:
                _os.remove(_config[name])
            else:
                raise FileExistsError('%s exists: ' % name + _config[name])
        _config[name] = src

    print("Update file %s" % name)
    if not _runtime:
        with open(_filepath, 'w') as _fp:
            _json.dump(_config, _fp)
    else:
        print("WARNING! The location of this file will be saved runtime only")


def add_from_dir(src, dest='', copy=True, overwrite=False, max_depth=-1):
    '''
    This function copy files from a src directory to dest directory,
    if dest is empty, then it creates a directory just under the edge_pydb
    package directory. The copy always assumes a topdown copy (i.e. from parent
    directory to child directories)

    Parameters
    ----------
    src: source directory to copy from
    dest: destination directory to copy to
    copy: if copy is false, files will not be copied, and instead the path of
            these files will be recorded in the _config.json
    overwrite: if file is at the destination, overwrite the file if true, else create
            and copy into a subdirectory data/ under the edge_pydb package directory
    max_depth: specify the depth the copy should perform.
            -1 means copy all directories
            0 means just under the root directory and do not go into subdirectories
    '''
    # this function assume a topdown copy
    if _runtime:
        print("WARNING! No sudo permission, take care, will break")
    dirname = _os.path.basename(src)
    # dirname = _os.path.abspath(_os.path.dirname(src))
    if not dest:
        dest = _ROOT + '/' + dirname

    if copy:
        if max_depth >= 0:
            src = _os.path.normpath(src)
            tmp = _walkthrough(src, max_depth)
            for k, v in tmp.items():
                # copy2 will not raise FileExist, but overwrite directly
                _shutil.copy2(v, dest)
                # print("here ", k, v)
        else:
            try:
                _shutil.copytree(src, dest)
            except FileExistsError:
                if overwrite:
                    _shutil.rmtree(dest)
                    _shutil.copytree(src, dest)
                else:
                    dest = _ROOT + '/data/' + dirname
                    _shutil.copytree(src, dest)
        updatefiles(dest)
    else:
        updatefiles(src, max_depth)

    if not _runtime:
        with open(_filepath, 'w') as _fp:
            _json.dump(_config, _fp)
    else:
        print("WARNING! The location of this file will be saved runtime only")

def getPath(file):
    '''get the hdf5 path'''
    h5f = _h5py.File(fetch(file), 'r')
    return [key for key in h5f.keys() if "__table_column_meta__" not in key]




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", 'r') as req:
    requirements = req.read()


#version_num = input("Input the new version number: ")
setuptools.setup(
    name="edge_pydb",
    version="1.4.0",  #version=version_num,
    author="Tony Wong",
    author_email="tonywong94@gmail.com",
    description="Python based database for CARMA EDGE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tonywong94/edge_pydb",
    packages=['edge_pydb'],
    package_data={
        'edge_pydb': [
            'dat_*/*/*.csv',
            'img_*/*/*.csv',
            'img_*/*.hdf5',
            '_config.json'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license='MIT',
    include_package_data=True,
    exclude_package_data={"": ['*/build/*']},
    install_requires=requirements,
)



import os
import datetime
import numpy as np
from scipy import stats
from matplotlib.patches import Circle
from matplotlib.collections import PatchCollection
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from edge_pydb.conversion import kewley01, kauffm03, cidfer10


def xy2hist(xarr, yarr, log=True, bins=[100,100]):
    '''
    Prepare a 2D density histogram from a scatterplot.
    Based on a response to a Stack Overflow question:
    stackoverflow.com/questions/49662964/density-scatter-plot-for-huge-dataset-in-matplotlib

    === Parameters ===
    xarr : numpy.array
        The x values in the scatter plot
    yarr : numpy.array
        The y values in the scatter plot
    log : boolean
        If True, take the log of input x and y, and of output density z
    bins : list of int
        The number of bins in x and y for the histogram

    === Returns ===
    x : numpy.array
        x values sorted by increasing density
    y : numpy.array
        y values sorted by increasing density
    z : numpy.array
        histogram count for each point
    hh : numpy.ndarray
        2D histogram values
    locx : numpy.array
        bin edges along first dimension
    locy : numpy.array
        bin edges along first dimension
    '''
    if log:
        x = np.log10(xarr)
        y = np.log10(yarr)
    else:
        x = xarr
        y = yarr
    # Histogram the data
    hh, locx, locy = np.histogram2d(x, y, bins=bins)
    # Get the bin value z for each (x,y) point
    z = np.array([hh[np.argmax(a<=locx[1:]),np.argmax(b<=locy[1:])] for a,b in zip(x,y)])
    # Plot the highest density points last
    idx = z.argsort()
    x, y, z = x[idx], y[idx], z[idx]
    if log:
        z = np.log10(z)
    return x, y, z, hh, locx, locy


def xy2binned(xarr, yarr, log=True, bins=20, range=None, yval='mean'):
    '''
    Prepare binned averages from a scatterplot.
    If log=True, averaging is done after taking the log of x and y.

    === Parameters ===
    xarr : numpy.array
        The x values in the scatter plot
    yarr : numpy.array
        The y values in the scatter plot
    log : boolean
        If True, take the log of input x and y
    bins : int
        The number of bins in x for the averaging
    range : (float, float)
        The lower and upper range of the bins. If not provided, uses (x.min(), x.max())
    yval : string
        'mean' (default) or 'median'

    === Returns ===
    xbin : numpy.array
        the centers (NOT edges) of the x bins
    ymean : numpy.array
        the binned y-values
    ystd : numpy.array
        standard deviation of y-values
    '''
    if log:
        good = (xarr>0) & (yarr>0)
        x = np.log10(xarr[good])
        y = np.log10(yarr[good])
    else:
        x = xarr
        y = yarr
    if yval == 'median':
        stat = lambda y: np.nanmedian(y)
        estat = lambda y: stats.iqr(y, scale='normal', nan_policy='omit')
    else:
        stat = lambda y: np.nanmean(y)
        estat = lambda y: np.nanstd(y)
    if bins > 0:
        ymean, xbinedge, _ = stats.binned_statistic(x, y,
            statistic=stat, bins=bins, range=range)
        ystd, xbinedge, _  = stats.binned_statistic(x, y,
            statistic=estat, bins=bins, range=range)
        xbin = 0.5*(xbinedge[1:]+xbinedge[:-1])
        return xbin, ymean, ystd
    else:
        return [0], [0], [0]


def dotpatch(x, y, imval, blank=None, dotsize=1, clipedge=True, pad=5, axes=None, **kwargs):
    '''
    Generate and plot a patch collection for a dot plot.

    === Parameters ===
    x : numpy.array
        The x values in the scatter plot (typically 'ix')
    y : numpy.array
        The y values in the scatter plot (typically 'iy')
    imval : numpy.array
        The z values which determine the dot colors
    blank : boolean array
        True values in this array are set to NaN
    dotsize : float
        The size of the filled circles
    clipedge : boolean
        True to derive a square bounding box around non-Nan values.
        False to show full image.
    pad : int
        Padding in pixels around edges if clipedge=True
    axes : matplotlib.axes
        Axes for plotting
    **kwargs :
        Additional arguments including vmin, vmax, colormap normalization

    === Returns ===
    img : matplotlib image object
        Plotted image
    xminmax : tuple
        [x.min(), x.max()]
    yminmax : tuple
        [y.min(), y.max()]
    '''
    # blank is a boolean array that sets certain pixels to NaN
    if blank is not None:
        imval[blank] = np.nan
    if axes is None:
        axes = plt.gca()
    if 'vmin' in kwargs:
        vmin = kwargs.pop('vmin')
    else:
        vmin = np.nanmin(imval)
    if 'vmax' in kwargs:
        vmax = kwargs.pop('vmax')
    else:
        vmax = np.nanmax(imval)
    patches = []
    for x1, y1 in zip(x,y):
        circle = Circle((x1, y1), dotsize)
        patches.append(circle)
    p = PatchCollection(patches, **kwargs)
    p.set_array(np.array(imval))
    p.set_clim([vmin, vmax])
    if clipedge:
        valid = ~np.isnan(imval)
        xspan = x[valid].max() - x[valid].min()
        yspan = y[valid].max() - y[valid].min()
        if xspan > yspan:
            xminmax = [x[valid].min()-pad, x[valid].max()+pad]
            yminmax = [y[valid].min()-pad-np.floor((xspan-yspan)/2), 
                       y[valid].min()+pad-np.floor((xspan-yspan)/2)+xspan]
        else:
            yminmax = [y[valid].min()-pad, y[valid].max()+pad]
            xminmax = [x[valid].min()-pad-np.floor((yspan-xspan)/2), 
                       x[valid].min()+pad-np.floor((yspan-xspan)/2)+yspan]
    else:
        xminmax = [x.min(), x.max()]
        yminmax = [y.min(), y.max()]
    img = axes.add_collection(p)
    return img, xminmax, yminmax


def imarrayplot(x, y, imval, blank=None, clipedge=True, pad=5, axes=None, **kwargs):
    '''
    Plot a pixel image from a data column.

    === Parameters ===
    x : numpy.array
        The x values in the image plot (typically 'ix')
    y : numpy.array
        The y values in the image plot (typically 'iy')
    imval : numpy.array
        The z values which determine the dot colors
    blank : boolean array
        True values in this array are set to NaN
    clipedge : boolean
        True to derive a square bounding box around non-Nan values.
        False to show full image.
    pad : int
        Padding in pixels around edges if clipedge=True
    axes : matplotlib.axes
        Axes for plotting
    **kwargs :
        Additional arguments including vmin, vmax, colormap normalization

    === Returns ===
    img : matplotlib image object
        Plotted image
    xminmax : tuple
        [x.min(), x.max()]
    yminmax : tuple
        [y.min(), y.max()]
    '''
    if blank is not None:
        imval[blank] = np.nan
    if axes is None:
        axes = plt.gca()
    xdim = len(np.unique(x))
    ydim = len(np.unique(y))
    imarray = np.reshape(imval, [ydim,xdim], order='F')
    if clipedge:
        valid = ~np.isnan(imarray)
        xflat = np.flatnonzero(np.any(valid, axis=0))
        yflat = np.flatnonzero(np.any(valid, axis=1))
        xspan = xflat.max() - xflat.min()
        yspan = yflat.max() - yflat.min()
        if xspan > yspan:
            xminmax = [xflat.min()-pad, xflat.max()+pad]
            yminmax = [yflat.min()-pad-np.floor((xspan-yspan)/2), 
                       yflat.min()+pad-np.floor((xspan-yspan)/2)+xspan]
        else:
            yminmax = [yflat.min()-pad, yflat.max()+pad]
            xminmax = [xflat.min()-pad-np.floor((yspan-xspan)/2), 
                       xflat.min()+pad-np.floor((yspan-xspan)/2)+yspan]
    else:
        xminmax = [0, xdim-1]
        yminmax = [0, ydim-1]
    img = axes.imshow(imarray, origin='lower', **kwargs)
    return img, xminmax, yminmax


def gridplot(edgetab=None, gallist=None, columnlist=None, 
            xrange=None, yrange=None, blank=None, plotstyle='image',
            cmap='jet', nx=7, ny=6, dotsize=1, pdfname=None, 
            vshow=False, clipedge=False, pad=5, verbose=False, **kwargs):
    '''
    Plot one column for multiple galaxies or multiple columns for 
    one galaxy on a grid.

    === Parameters ===
    edgetab : EdgeTable
        Table containing the galaxies and data to plot
    gallist : string or list of strings
        List of galaxy names to plot.  If plotting multiple columns,
        only one galaxy should be given.
    columnlist : string or list of strings
        List of columns in the table to plot.  If plotting multiple
        galaxies, only one column should be given.
    xrange : tuple of float
        x limits applied to each panel (pixels)
    yrange : tuple of float
        y limits applied to each panel (pixels)
    blank : boolean array with same length as 'column'
        True values in this array are set to NaN
    plotstyle : string
        'dot' for dot plot, 'image' for pixel image
    cmap : string
        name of color map
    nx : int
        number of sub-panels in x direction
    ny : int
        number of sub-panels in y direction
    dotsize : float
        size of plot symbol for dot plot
    pdfname : string
        name of output PDF file, otherwise plot to screen
    vshow : boolean
        True to show vmin and vmax in plotting window
    clipedge : boolean
        True to derive a square bounding box around non-Nan values.
        False to use show full image.  Overridden by xrange, yrange.
    pad : int
        Padding in pixels around edges if clipedge=True
    **kwargs :
        Additional arguments including vmin, vmax, colormap normalization
    '''

    if gallist is None and columnlist is None:
        raise TypeError('Either gallist or columnlist must be provided!')
    if isinstance(gallist, str):
        gallist = [gallist]
    if isinstance(columnlist, str):
        columnlist = [columnlist]

    # Plot mode: multiple galaxies or multiple columns
    if columnlist is not None and len(columnlist) == 1:
        mode = 'onecol'
        # Plot all galaxies by default
        if gallist is None:
            gallist = list(np.unique(edgetab['Name']))
        print('Plotting column',columnlist[0],'for',len(gallist),'galaxies')
        pagelist = gallist
    elif gallist is not None and len(gallist) == 1:
        mode = 'onegal'
        # Plot all non-coordinate columns by default
        if columnlist is None:
            columnlist = edgetab.colnames
            for key in ['Name', 'ix', 'iy', 'ra_abs', 'dec_abs', 'ra_off', 'dec_off']:
                if key in columnlist:
                    columnlist.remove(key)
        print('Plotting',len(columnlist),'columns for galaxy',gallist[0])
        pagelist = columnlist
    else:
        raise ValueError('Specify either one galaxy or one column to plot')

    # Get default axis limits
    pages = int(np.ceil(float(len(pagelist)) / (nx*ny)))
    if pdfname is not None:
        pp = PdfPages(pdfname)

    for num in range(0,pages):
        aa = nx*ny*num
        bb = nx*ny+aa
        thispage = pagelist[aa:bb]
        fig = plt.figure(figsize=(18,14))
        if not verbose:
            print('Plotting', thispage[0], 'to', thispage[-1])

        for i in range(0,len(thispage)):
            if mode == 'onecol':
                gname = thispage[i]
                column = columnlist[0]
                label = gname
            else:
                gname = gallist[0]
                column = thispage[i]
                label = column
            if verbose:
                print('Plotting', thispage[i])
            ax = plt.subplot(ny,nx,i+1)
            galtab = (edgetab['Name'] == gname)
            if blank is not None:
                galblank = blank[galtab]
            else:
                galblank = None
            if plotstyle == 'dot':
                img, xlims, ylims = dotpatch(edgetab[galtab]['ix'], 
                                            edgetab[galtab]['iy'],
                                            edgetab[galtab][column], 
                                            blank=galblank, clipedge=clipedge,
                                            pad=pad, dotsize=dotsize, cmap=cmap, 
                                            axes=ax, **kwargs)
            else:
                img, xlims, ylims = imarrayplot(edgetab[galtab]['ix'], 
                                            edgetab[galtab]['iy'],
                                            edgetab[galtab][column], 
                                            blank=galblank, clipedge=clipedge,
                                            pad=pad, cmap=cmap, axes=ax, **kwargs)        
            if xrange is None:
                ax.set_xlim(xlims)
                if i == 0:
                    print(label, "Default x limits used:",xlims)
            else:
                ax.set_xlim(xrange)
            if yrange is None:
                ax.set_ylim(ylims)
                if i == 0:
                    print(label, "Default y limits used:",ylims)
            else:
                ax.set_ylim(yrange)
            ax.set_aspect('equal')
            ax.xaxis.set_ticks([])
            ax.yaxis.set_ticks([])
            plt.text(0.04,0.92,label,ha='left',va='center',transform=ax.transAxes,
               bbox=dict(facecolor='white', edgecolor='none', pad=1))
            if vshow:
                vminmax = img.get_clim()
                if vminmax[1] < 1:
                    labelstr = '[{:.3f} .. {:.3f}]'.format(vminmax[0],vminmax[1])
                else:
                    labelstr = '[{:.2f} .. {:.2f}]'.format(vminmax[0],vminmax[1])
                plt.text(0.04,0.07,labelstr,ha='left',va='center',
                    transform=ax.transAxes, bbox=dict(facecolor='none',edgecolor='none'))
        fig.subplots_adjust(hspace=0.05)
        fig.subplots_adjust(wspace=0.05)
        if pdfname is not None:
            pp.savefig(bbox_inches = 'tight', pad_inches=0.1)
            plt.close()
        else:
            plt.show()

    if pdfname is not None:
        d = pp.infodict()
        d['Title'] = 'EDGE Gallery'
        d['Author'] = os.getlogin()
        d['CreationDate'] = datetime.datetime.today()
        pp.close()
    return


def plot_uncertainty_ellipse(xval_n, xval_s, yval_n, yval_s, indices, x_arr, save_to=''):
    '''
    parameters
    xval_n, yval_u : list of nominal values of coordinates (pixels) 
    xval_s, yval_s : list of standard deviation of coordinates (pixels)
    indices : indices of the list of coordinates to plot with
    save_to: file to save the plot to, optional
    '''
    import matplotlib.pyplot as plt
    from matplotlib.patches import Ellipse
    plt.figure(figsize=(8,8))
    ax = plt.gca()
    for i in indices:
        ax.add_patch(Ellipse(xy=(xval_n[i], yval_n[i]),
                            width=xval_s[i], height=yval_s[i],
                            edgecolor='red', facecolor='none'))
    plt.plot(x_arr, kewley01(x_arr), 'k-.', label="Kewley")
    plt.plot(x_arr, kauffm03(x_arr), 'k--', label="Kauffmann")
    plt.plot(x_arr, cidfer10(x_arr), 'k-', label="Cidfer")
    if save_to:
        plt.savefig(save_to)
    plt.show()



import numpy as np
from astropy.io import fits
from astropy.table import Table, Column
from astropy import units as u
from astropy.coordinates import SkyCoord
from scipy.optimize import fsolve
from scipy import ndimage
from scipy.stats import multivariate_normal as ndNormal

try:
    from uncertainties import unumpy, umath
except (ImportError, ModuleNotFoundError) as error:
    print(error.__class__.__name__ + ": " + str(error))
    print("Could not import uncertainties package, \
        please try to install it or do not input the error as argument.")
except Exception as exception: 
    print(exception.__class__.__name__ + ": " + str(exception))

STAR_FORMING = -1
INTERMEDIATE = 0
LINER = 1
SEYFERT = 2

def gc_polr(ra, dec, ra_gc, dec_gc, pa, inc, reject=89):
    '''
    Returns galactocentric polar coordinates (radius in arcsec, 
    azimuthal angle in degrees from receding majaxis).

    === Parameters ===
    ra : float or numpy.array
        The RA coordinate(s) to transform, in degrees
    dec : float or numpy.array
        The DEC coordinate(s) to transform, in degrees
    ra_gc : float
        The RA of the galaxy center, in degrees
    dec_gc : float
        The DEC of the galaxy center, in degrees
    pa : float
        The position angle of the galaxy's receding major axis, in degrees
    inc : float
        The inclination of the galaxy disk, in degrees (0 = face-on)
    reject : float
        Return NaNs for inclinations larger than this (default=89)

    === Returns ===
    pair of floats or numpy.array [radius, azang]
    '''
    ctr = SkyCoord(ra_gc, dec_gc, unit="deg")
    pos = SkyCoord(ra, dec, unit="deg")
    # Polar vector in sky plane 
    t_sky = ctr.position_angle(pos).degree
    r_sky = ctr.separation(pos).arcsecond
    # Convert to galaxy plane
    x_sky = r_sky * np.cos(np.radians(t_sky - pa))
    y_sky = r_sky * np.sin(np.radians(t_sky - pa))
    x_gal = x_sky
    y_gal = y_sky/np.cos(np.radians(inc))
    # Reject very high inclinations (cannot be deprojected)
    y_gal[inc>reject] = np.nan
    radius = np.sqrt(x_gal**2 + y_gal**2)
    azang  = np.degrees(np.arctan2(y_gal, x_gal))
    return radius, azang


def uarray_to_list(target):
    '''
    Break array of uncertainties values into 2 separate arrays.

    === Parameters ===
    target : unumpy.uarray
        The array of values with uncertainties

    === Returns ===
    retval[0] -> nominal values
    retval[1] -> standard deviation
    '''
    return [unumpy.nominal_values(target), unumpy.std_devs(target)]


def get_AHa(flux_ha, flux_hb, log10):
    '''
    Returns Halpha extinction in magnitudes given flux_ha and flux_hb.

    === Parameters ===
    flux_ha : astropy.table.Column
        The H-alpha flux values
    flux_hb : astropy.table.Column
        The H-beta flux values
    log10 : function
        The log10 function, normally np.log10 but use unumpy.log10 for error propagation.

    === Returns ===
    numpy.ndarray representing AHa in magnitudes
    '''
    A_Ha = flux_ha.data * np.nan
    if flux_hb is not None:
        # Need positive flux to take the log.
        good = (flux_ha > 0) & (flux_hb > 0)
        # Extinction curve from Cardelli+ (1989).
        K_Ha = 2.53
        K_Hb = 3.61
        # Get A_Ha using Eq(1) from Catalan-Torrecilla+ (2015). 
        A_Ha[good] = K_Ha/(-0.4*(K_Ha-K_Hb)) * log10((flux_ha[good]/flux_hb[good])/2.86)
    return A_Ha


def sfr_ha(flux_ha, flux_hb=None, e_flux_ha=None, e_flux_hb=None, 
            name='sigsfr', column=True, imf='kroupa'):
    '''
    Convert Halpha intensity to SFR surface density, optionally
    with extinction estimates and corrections (if flux_hb is provided).
    Note that both e_flux_ha and e_flux_hb have to be not None
    in order to propagate the error.  Otherwise the SFR is computed
    without dust correction (if flux_hb=None) or without error
    estimation. 

    === Parameters ===
    flux_ha : astropy.table.Column
        The H-alpha flux values
    flux_hb : astropy.table.Column
        The H-beta flux values.  If not provided, no dust correction is made.
    e_flux_ha : astropy.table.Column
        Errors in H-alpha flux values.  If not provided, error propagation skipped.
    e_flux_hb : astropy.table.Column
        Errors in H-beta flux values.
    name : string
        Name of the output Column; error column will be prepended with 'e_'
    column : boolean
        True to return astropy Column objects, otherwise return arrays
    imf : string
        'salpeter' to scale by 1.51 (default uses Kroupa IMF)

    === Returns ===
    several columns depending on input (see code)
    '''
    # input line flux is actually flux per arcsec^2
    sterad = (u.sr/u.arcsec**2).decompose().scale # 206265^2
    
    # Calzetti+(2010ApJ...714.1256C), Kroupa IMF, 1 Gyr old pop
    lumcon = 5.45e-42 * (u.solMass/u.yr) / (u.erg/u.s)
    if imf == 'salpeter':
        lumcon *= 1.51

    # Case with error estimation
    if e_flux_ha is not None:
        u_flux_ha = unumpy.uarray(flux_ha, e_flux_ha)
        sig_sfr = (lumcon * 4*np.pi * u_flux_ha * sterad).to(u.solMass/(u.pc**2*u.Gyr))
        if flux_hb is None:   # no Balmer decrement
            sig_sfr_out = uarray_to_list(sig_sfr.value)
            if column:
                return Column(sig_sfr_out[0], name=name, dtype='f4', unit=sig_sfr.unit,
                    description='SFR surface density no extinction'), \
                    Column(sig_sfr_out[1], name='e_'+name, dtype='f4', unit=sig_sfr.unit,
                    description='error of uncorrected SFR surface density')
            else:
                return sig_sfr_out[0], sig_sfr_out[1]
        else:   # with Balmer decrement
            if e_flux_hb is None:
                e_flux_hb = np.zeros_like(flux_hb)
            u_flux_hb = unumpy.uarray(flux_hb, e_flux_hb)
            A_Ha = get_AHa(u_flux_ha, u_flux_hb, unumpy.log10)
            sig_sfr = sig_sfr * 10**(0.4*A_Ha)
            # * sig_sfr is a astropy Quantity, and A_Ha is still a Column
            sig_sfr_out = uarray_to_list(sig_sfr.value)
            A_Ha_out = uarray_to_list(A_Ha.data)
            if column:
                return Column(sig_sfr_out[0], name=name, dtype='f4', unit=sig_sfr.unit,
                    description='BD corrected SFR surface density'), \
                    Column(A_Ha_out[0], name=name.replace('sigsfr','AHa'), 
                    dtype='f4', unit='mag', description='Ha extinction from BD'), \
                    Column(sig_sfr_out[1], name='e_'+name, dtype='f4', unit=sig_sfr.unit,
                    description='error of BD corrected SFR surface density'), \
                    Column(A_Ha_out[1], name='e_'+name.replace('sigsfr','AHa'), 
                    dtype='f4', unit='mag', description='error of Ha extinction from BD')
            else:
                return sig_sfr_out[0], A_Ha_out[0], sig_sfr_out[1], A_Ha_out[1]

    # Case with no error estimation
    else:
        sig_sfr = (lumcon * 4*np.pi * flux_ha * sterad).to(u.solMass/(u.pc**2*u.Gyr))
        if flux_hb is None:   # no Balmer decrement
            if column:
                return Column(sig_sfr, name=name, dtype='f4',
                    description='SFR surface density no extinction')
            else:
                return sig_sfr
        else:   # with Balmer decrement
            A_Ha = get_AHa(flux_ha, flux_hb, np.log10)
            sig_sfr = sig_sfr * 10**(0.4*A_Ha)
            if column:
                return Column(sig_sfr, name=name, dtype='f4',
                    description='BD corrected SFR surface density'), \
                    Column(A_Ha, name=name.replace('sigsfr','AHa'), dtype='f4', 
                    unit='mag', description='Ha extinction from BD')
            else:
                return sig_sfr, A_Ha


def msd_co(sb_co, alphaco=4.3, name='sigmol'):
    '''
    Convert CO intensity to H_2 (+ He) mass surface density.

    === Parameters ===
    sb_co : numpy.array or astropy.table.Column
        CO integrated intensity values, in K km/s
    alphaco : float
        CO to H2 conversion factor, from K km/s to Msol/pc2
    name : string
        The name of the output Column, if input is a Column

    === Returns ===
    numpy.array or astropy.table.Column
    '''
    convfac = alphaco * (u.solMass/u.pc**2) / (u.K*u.km/u.s)
    sig_mol = (convfac*sb_co).to(u.solMass/u.pc**2)
    if isinstance(sb_co, Column):
        return Column(sig_mol, name=name, dtype='f4',
                      description='mol gas surface density')
    else:
        return sig_mol
    

def stmass_pc2(stmass_as2, dz=None, dist=10*u.Mpc, name='sigstar'):
    '''
    Convert units for stellar surface density to Msol/pc2, optionally 
    applying dezonification image.
    
    === Parameters ===
    stmass_as2 : numpy.array or astropy.table.Column
        stellar surface density in Pipe3D units
    dz : numpy.array or astropy.table.Column
        dezonification image from Pipe3D (optional)
    dist : float or astropy.Quantity
        The distance of the galaxy, assumed to be in Mpc if no units
    name : string
        The name of the output Column, if input is a Column

    === Returns ===
    numpy.array or astropy.table.Column
    '''
    # Assume Mpc units if not given
    try:
        unit = dist.unit
    except:
        dist = dist * u.Mpc
    sterad = (u.sr/u.arcsec**2).decompose().scale   # 206265^2
    pxarea = (dist**2/sterad).to(u.pc**2)

    def convert_stmass(stmass_in, dz):
        stmass_in[~np.isfinite(stmass_in)] = np.nan
        stmass_pc2 = 10**stmass_in * u.solMass / pxarea
        if dz is not None:
            stmass_pc2 *= np.array(dz)
        stmass_pc2[~np.isfinite(stmass_pc2)] = np.nan
        return stmass_pc2

    if isinstance(stmass_as2, Column):
        stmass_ary = np.array(stmass_as2)
        stmass_pc2 = convert_stmass(stmass_ary, dz)
        return Column(stmass_pc2, name=name, dtype='f4',
                      description='stellar mass surface density')
    else:
        stmass_pc2 = convert_stmass(stmass_as2, dz)
        return stmass_pc2


def findIntersec(fun1, fun2, x0):
    return fsolve(lambda x:fun1(x)-fun2(x), x0)


def kewley01(nii):
    # Eq. 5 of 2001ApJ...556..121K
    return 1.19 + 0.61/(nii - 0.47) 


def kauffm03(nii):
    # Eq. 1 of 2003MNRAS.346.1055K
    return 1.30 + 0.61/(nii - 0.05)


def cidfer10(nii):
    # Eq. 3 of 2010MNRAS.403.1036C
    return 0.48 + 1.01*nii    


def bpt_region(n2ha, o3hb, good=True):
    '''
    Determine BPT classification based on [NII]/Ha and [OIII]/Hb.  Does 
    not impose a requirement on EW(Ha).

    === Parameters ===
    n2ha : numpy.ndarray
        log10([NII]/Halpha)
    o3hb : numpy.ndarray
        log10([OIII]/Hbeta)
    good : boolean
        mask to apply to output arrays    

    === Returns ===
    boolean index arrays for SF, intermediate, LINER, and Seyfert.
    '''
    # These are constant values determined by the curve equations
    kewley_start = findIntersec(kewley01, kauffm03, -1)[0]  # -1.2805
    cidfer_start = findIntersec(kewley01, cidfer10, 0)[0]   # -0.1993
    kewley_end = findIntersec(kewley01, lambda x: -4, 0)[0] #  0.352466
    kauffm_end = findIntersec(kauffm03, lambda x: -4, 0)[0] # -0.06509
    # Star forming: below Kauffmann line
    sf = (n2ha < kauffm_end) & (o3hb < kauffm03(n2ha))
    # Intermediate: below Kewley line and not star-forming
    inter = (~sf) & (n2ha < kewley_end) & (o3hb < kewley01(n2ha))
    # LINER: above Kewley line and below Cid Fernandes line
    liner = (~sf) & (~inter) & (o3hb < cidfer10(n2ha))
    # Seyfert: above Kewley line and above Cid Fernandes line
    seyfert = (~sf) & (~inter) & (~liner) & good
    retval = np.full_like(n2ha, fill_value=np.nan)
    retval[sf] = STAR_FORMING
    retval[inter] = INTERMEDIATE
    retval[liner] = LINER
    retval[seyfert] = SEYFERT
    return retval


def bpt_prob(n2ha_u, o3hb_u, bpt_type, grid_size=5):
    '''
    Calculate the probability that a measurement with uncertainty is in a particular
    region of the BPT diagram.  This is done by constructing a grid and measuring the
    normalized Gaussian area within the chosen BPT region.

    === Parameters ===
    n2ha_u : ufloat
        log10([NII]/Halpha) and uncertainty
    o3hb_u : ufloat
        log10([OIII]/Hbeta) and uncertainty
    bpt_type : float
        BPT type to measure prob for, -1=SF, 0=intermediate, 1=LINER, 2=Seyfert
    grid_size : float
        The size of the square grid where normal dist constructed
    
    === Returns ===
    probability that measurement is in the specified region
    '''
    x = unumpy.nominal_values(n2ha_u)
    y = unumpy.nominal_values(o3hb_u)
    x_std = unumpy.std_devs(n2ha_u)
    y_std = unumpy.std_devs(o3hb_u)
    x_arr, y_arr = np.meshgrid( np.linspace(x - x_std, x + x_std, grid_size),
                                np.linspace(y - y_std, y + y_std, grid_size) )
    pos = np.dstack((x_arr, y_arr))
    grid = np.zeros((grid_size, grid_size))  
    grid = bpt_region(x_arr, y_arr)
    gauss2d = ndNormal(mean=(x, y), cov=(x_std**2, y_std**2))
    pdf = gauss2d.pdf(pos)
    normal_prob = np.zeros((grid_size, grid_size))
    delta_x = x_arr[0][1] - x_arr[0][0]
    delta_y = y_arr[1][0] - y_arr[0][0]
    total = 0
    # the integrand
    normal_prob = pdf * delta_x * delta_y
    # Normalization constant, count how many cells we have
    total = np.sum(normal_prob)
    # do the sum reduction
    return ndimage.sum(normal_prob * 1/total, grid, index=bpt_type)



def bpt_type(fluxtab, ext='', name='BPT', sf=True, prob=False, grid_size=5):
    '''
    Adds BPT classification to the flux_elines table. BPT==-1 means in the
    star-forming region of the diagram.  An additional column SF_BPT is
    True when BPT==-1 and Halpha EW > 6. 
    This function should be run before ZOH_M13.
    For more information see Husemann et al. (2013A&A...549A..87H) Figure 7.

    === Parameters ===
    fluxtab : astropy.Table
        flux_elines table extracted from Pipe3D output
    ext : string
        suffix for selected column names, e.g. '_rg' or '_sm'
    name : string
        name of the output column
    sf : boolean
        True to calculate the SF_BPT column using Halpha EW.
    prob : boolean
        True to calculate the BPT probabilities as an additional column
    grid_size : float
        The size of the square grid where BPT probabilities constructed
    
    === Returns ===
    two or three astropy.table.Column [BPT, SF_BPT, p_BPT]
    '''
    flux_nii  = fluxtab['flux_[NII]6583'+ext]
    flux_oiii = fluxtab['flux_[OIII]5007'+ext]
    flux_ha   = fluxtab['flux_Halpha'+ext]
    flux_hb   = fluxtab['flux_Hbeta'+ext]

    good = (flux_nii>1e-5) & (flux_oiii>1e-5) & (flux_ha>1e-5) & (flux_hb>1e-5)
    n2ha = np.full(len(flux_nii), np.nan)
    n2ha[good] = np.log10(flux_nii[good])  - np.log10(flux_ha[good])
    o3hb = np.full(len(flux_oiii), np.nan)
    o3hb[good] = np.log10(flux_oiii[good]) - np.log10(flux_hb[good])   

    BPT = bpt_region(n2ha, o3hb, good)
    bpt_col = Column(BPT, name=name, dtype='f4', format='.1f',
                description='BPT type (-1=SF 0=inter 1=LINER 2=Sy)')
    if sf:
        ew_ha = fluxtab['EW_Halpha'+ext]
        bpt_sf = (BPT == -1) & (abs(ew_ha)> 6.0)
        bpt_sfcol = Column(bpt_sf, name='SF_' + name, dtype='?',
                    description='True if star forming (BPT=-1 and EW_Ha>6)')

    if prob:
        BPT_prob = np.full(len(n2ha), np.nan)
        eN2 = fluxtab['e_flux_[NII]6583'+ext]
        eO3 = fluxtab['e_flux_[OIII]5007'+ext] 
        eHa = fluxtab['e_flux_Halpha'+ext]
        eHb = fluxtab['e_flux_Hbeta'+ext]
        print("Calculating the probability")
        Ha_u = unumpy.uarray(np.array(flux_ha), np.array(eHa))
        Hb_u = unumpy.uarray(np.array(flux_hb), np.array(eHb))
        N2_u = unumpy.uarray(np.array(flux_nii), np.array(eN2))
        O3_u = unumpy.uarray(np.array(flux_oiii), np.array(eO3))
        t1 = [unumpy.log10(N2_u[good][i]) - unumpy.log10(Ha_u[good][i]) for i in range(len(Ha_u[good]))]
        t2 = [unumpy.log10(O3_u[good][i]) - unumpy.log10(Hb_u[good][i]) for i in range(len(Hb_u[good]))]
        n2ha_u = unumpy.uarray(np.full(len(Ha_u), np.nan), np.full(len(Ha_u), np.nan))
        o3hb_u = unumpy.uarray(np.full(len(Hb_u), np.nan), np.full(len(Hb_u), np.nan))
        n2ha_u[good] = unumpy.uarray(unumpy.nominal_values(t1), unumpy.std_devs(t1))
        o3hb_u[good] = unumpy.uarray(unumpy.nominal_values(t2), unumpy.std_devs(t2)) 
        print("Working on SFR")
        for i in np.where(BPT == STAR_FORMING)[0]:
            BPT_prob[i] = bpt_prob(n2ha_u[i], o3hb_u[i], STAR_FORMING, grid_size)
        print("Working on intermediate region, composite")
        for i in np.where(BPT == INTERMEDIATE)[0]:
            BPT_prob[i] = bpt_prob(n2ha_u[i], o3hb_u[i], INTERMEDIATE, grid_size)
        print("Working on liner")
        for i in np.where(BPT == LINER)[0]:
            BPT_prob[i] = bpt_prob(n2ha_u[i], o3hb_u[i], LINER, grid_size)
        print("Working on Seyfert")
        for i in np.where(BPT == SEYFERT)[0]:
            BPT_prob[i] = bpt_prob(n2ha_u[i], o3hb_u[i], SEYFERT, grid_size)
        prob_col = Column(BPT_prob, name='p_'+name, dtype='f4', description='BPT probability')

    if sf and prob:
        return bpt_col, bpt_sfcol, prob_col
    elif sf:
        return bpt_col, bpt_sfcol
    else:
        return bpt_col


# Metallicity derived from Marino+13 calibration.
# use method='o3n2' or method='n2'
# Require star-forming in BPT diagram.
# Input is a flux_elines table.
def ZOH_M13(fluxtab, ext='', method='o3n2', name='ZOH', err=True):
    '''
    Adds gas-phase metallicity from Marino+13 calibration to flux_elines table.
    This function should be run after bpt_type since it requires spaxel to be
    star-forming in the BPT diagram.
    Both O3N2 and N2 methods are supported.

    === Parameters ===
    fluxtab : astropy.Table
        flux_elines table extracted from Pipe3D output
    ext : string
        suffix for selected column names, e.g. '_rg' or '_sm'
    method : string
        choose 'o3n2' (default) or 'n2'
    err : boolean
        True to calculate uncertainty as an additional column
    
    === Returns ===
    one or two astropy.table.Column [ZOH, e_ZOH]
    '''
    N2F = fluxtab['flux_[NII]6583'+ext]
    O3F = fluxtab['flux_[OIII]5007'+ext]
    HaF = fluxtab['flux_Halpha'+ext]
    HbF = fluxtab['flux_Hbeta'+ext]
    BPT_sf = fluxtab['SF_BPT'+ext]

    if method == 'o3n2':
        good = (N2F>0) & (O3F>0) & (HaF>0) & (HbF>0) & BPT_sf
    elif method == 'n2':
        good = (N2F>0) & (HaF>0) & BPT_sf
    else:
        raise Exception('Method {} is not recognized'.format(method))
    nelt = len(N2F)

    if err == False:
        O3N2 = np.full(nelt, np.nan)
        O3N2[good] = (np.log10(O3F[good]) - np.log10(HbF[good]) 
                    - (np.log10(N2F[good]) - np.log10(HaF[good])))                    
        N2 = np.full(nelt, np.nan)
        N2[good] = np.log10(N2F[good]) - np.log10(HaF[good])                   
    else:
        from uncertainties import ufloat
        import uncertainties.unumpy as unp 

        uN2F = unp.uarray(N2F, fluxtab['e_flux_[NII]6583'+ext])
        uO3F = unp.uarray(O3F, fluxtab['e_flux_[OIII]5007'+ext])
        uHaF = unp.uarray(HaF, fluxtab['e_flux_Halpha'+ext])
        uHbF = unp.uarray(HbF, fluxtab['e_flux_Hbeta'+ext])
        O3N2 = unp.uarray(np.full(nelt, np.nan),np.full(nelt, np.nan))
        O3N2[good] = (unp.log10(uO3F[good]) - unp.log10(uHbF[good]) 
                    - (unp.log10(uN2F[good]) - unp.log10(uHaF[good])))
        N2 = unp.uarray(np.full(nelt, np.nan),np.full(nelt, np.nan))
        N2[good] = unp.log10(uN2F[good]) - unp.log10(uHaF[good])

    if method == 'o3n2':
        ZOH_M13 = 8.533 - 0.214 * O3N2 # Eq(2) from Marino+2013
    else:
        ZOH_M13 = 8.743 + 0.462 * N2   # Eq(4) from Marino+2013

    desc = '12+log(O/H) using {} method in Marino+13'.format(method)
    if err == False: 
        return Column(ZOH_M13, name=name, unit='dex', dtype='f4', description=desc)
    else:            
        return (Column(unp.nominal_values(ZOH_M13), name=name, dtype='f4',
                       unit='dex', description=desc), 
                Column(unp.std_devs(ZOH_M13), name='e_'+name, dtype='f4',
                       unit='dex', description='error in '+desc))

